@@ -447,7 +447,7 @@ giop_mainloop_handle_input (GIOChannel     *source,
 	LINK_MUTEX_LOCK (tdata->lock);
 	while (!giop_thread_queue_empty_T (tdata)) {
 		LINK_MUTEX_UNLOCK (tdata->lock);
-		giop_thread_queue_process (NULL);
+		giop_thread_queue_process (tdata);
 		LINK_MUTEX_LOCK (tdata->lock);
 	}
 	LINK_MUTEX_UNLOCK (tdata->lock);
@@ -515,6 +515,7 @@ giop_init (gboolean threaded, gboolean blank_wire_data)
 
 		if (pipe (corba_wakeup_fds) < 0) /* cf. g_main_context_init_pipe */
 			g_error ("Can't create CORBA main-thread wakeup pipe");
+		fcntl (WAKEUP_WRITE, F_SETFL, O_NONBLOCK);
 
 		giop_main_source = link_source_create_watch (
 			g_main_context_default (), WAKEUP_POLL,
@@ -542,32 +543,24 @@ wakeup_mainloop (void)
 	char c = 'A'; /* magic */
 	int  res;
 	while ((res = write (WAKEUP_WRITE, &c, sizeof (c))) < 0  &&
-	       (errno == EAGAIN || errno == EINTR));
+	       errno == EINTR );
+	if (res < 0 && errno == EAGAIN)
+		return;
 	if (res < 0)
-		g_error ("Failed to write to GIOP wakeup socket %d 0x%x(%d) (%d)",
-			 res, errno, errno, WAKEUP_WRITE);
+		g_warning ("Failed to write to GIOP mainloop wakeup "
+			   "pipe %d 0x%x(%d) (%d)",
+			   res, errno, errno, WAKEUP_WRITE);
 }
 
 void
-giop_incoming_signal_T (GIOPThread *tdata)
+giop_incoming_signal_T (GIOPThread *tdata, GIOPMsgType t)
 {
 	g_cond_signal (tdata->incoming);
 
-	if (tdata->wake_context)
+	if (t != GIOP_REPLY && tdata->wake_context)
 		wakeup_mainloop ();
 }
 
-void
-giop_thread_push_recv (GIOPMessageQueueEntry *ent)
-{
-	g_return_if_fail (ent != NULL);
-	g_return_if_fail (ent->buffer != NULL);
-	g_return_if_fail (ent->src_thread != NULL);
-
-	/* someone already waiting on the stack */
-	giop_incoming_signal_T (ent->src_thread);
-}
-
 void
 giop_invoke_async (GIOPMessageQueueEntry *ent)
 {
@@ -590,7 +583,7 @@ giop_invoke_async (GIOPMessageQueueEntry *ent)
 
 		buf = NULL;
 		tdata->async_ents = g_list_prepend (tdata->async_ents, ent);
-		giop_incoming_signal_T (tdata);
+		giop_incoming_signal_T (tdata, GIOP_REQUEST);
 		
 		g_mutex_unlock (tdata->lock); /* ent_unlock */
 	}
@@ -662,6 +655,7 @@ first_valid_request (GIOPThread *tdata, gboolean *no_policy)
 	}
 
 	*no_policy = FALSE;
+	policy = g_queue_peek_head (tdata->invoke_policies);
 
 	for (l = tdata->request_queue; l; l = l->next) {
 		int i;
@@ -738,7 +732,7 @@ giop_thread_queue_process (GIOPThread *tdata)
 		}
 	}
 
-	dprintf (MESSAGES, "Queue pop %p, %p, %d", ent, qe);
+	dprintf (MESSAGES, "Queue pop %p, %p, %d", ent, qe, no_policy);
 	
 	LINK_MUTEX_UNLOCK (tdata->lock); /* ent_unlock */
 
@@ -786,7 +780,7 @@ giop_thread_request_push (GIOPThread *tdata,
 	LINK_MUTEX_LOCK (tdata->lock);
 
 	tdata->request_queue = g_list_append (tdata->request_queue, qe);
-	giop_incoming_signal_T (tdata);
+	giop_incoming_signal_T (tdata, GIOP_REQUEST);
 
 	LINK_MUTEX_UNLOCK (tdata->lock);
 }
